/*
* Shutter - Friend List Component
*
* Copyright (C) 2004 luma <stubma@163.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
package edu.tsinghua.swt.layout;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Layout;

import edu.tsinghua.swt.internal.layout.DefaultKeyComparator;
import edu.tsinghua.swt.internal.layout.SortDataIndexComparator;

/**
 * <pre>
 * My own layout, it's a layout with sort function, composite with this layout will place
 * children one by one. One child can be added to the last or inserted in
 * the middle position.
 * Controls in this layout can be sorted on "key" field in control's SortData.
 * If you don't choose sorting controls, the controls will be placed by its index,
 * further more, you can specify a fixed location to a control, such as a fixed index
 * or adhere to other control
 * </pre>
 * 
 * @author 
 */
public final class SortLayout extends Layout {
	// ListData array
	private Vector list;
	// layout variable
	public int marginWidth;
	public int marginHeight;
	public int numColumns;
	public boolean makeColumnsEqualWidth;
	public boolean makeRowsEqualHeight;
	public boolean sort;
	public boolean ascendent;
	public int horizontalSpacing;
	public int verticalSpacing;
	// key comparator
	public Comparator keyComparator;
	// sort mask, if ascendent is true, it's 1, otherwise is -1
	private int sortMask;
	// SortData comparator
	private SortDataIndexComparator comparator;
	// expandable rows and cols
	private int[] expandableRows, expandableCols;
	// max height of every row, not including margin
	private int[] rowMaxHeight;
	// max width of every column, not including margin
	private int[] colMaxWidth;
	
	/**
	 * constructor
	 */
	public SortLayout()
	{
		this(1, false, true);
	}
	
	/**
	 * constructor
	 * @param numColumns
	 */
	public SortLayout(int numColumns) {
		this(numColumns, false, true);
	}
	
	/**
	 * constructor
	 * @param numColumns
	 * @param sort
	 */
	public SortLayout(int numColumns, boolean sort) {
		this(numColumns, sort, true);
	}
	
	/**
	 * constructor
	 * @param numColumns
	 * @param sort true if we need sort
	 */
	public SortLayout(int numColumns, boolean sort, boolean ascendent) {
		this.numColumns = numColumns;
		this.sort = sort;
		this.ascendent = ascendent;
		makeColumnsEqualWidth = false;
		marginWidth = marginHeight = 5;
		horizontalSpacing = verticalSpacing = 5;
		list = new Vector();
		comparator = new SortDataIndexComparator();
		keyComparator = new DefaultKeyComparator();
	}
	
	// create list layout data array from composite
	private void createList(Composite composite) {
		// set sort mask
		sortMask = ascendent?1:-1;
		// fixed control array
		Vector adhereToIndexs = new Vector();
		Vector adhereToControls = new Vector();
		// begin create list
		Control[] children = composite.getChildren();
		for(int i=0;i<children.length;i++) {
			// get layout data, if null, set a default
			SortData sd = (SortData)children[i].getLayoutData();
			if(sd==null) {
				sd = new SortData();
				children[i].setLayoutData(sd);
			}
			// set control's index in children array
			sd.childIndex = i;
			// calculate control's layout index and put it into list
			//    if the index is larger than list size, put it as the
			//    last one, but this should only occur when the sort is
			//    false, if sort is true, layout sequence will be determined
			//    by ListData's key field, if key field is null, place
			//    it to the last
			//    if this control is adhered to something, the ignore sort flag
			if(sd.adhereTo==SortData.INDEX)
				adhereToIndexs.add(sd);
			else if(sd.adhereTo==SortData.CONTROL)
				adhereToControls.add(sd);
			else if(sort) {
				if(sd.key==null) list.add(sd);
				else {
					int size = list.size();
					int j;
					for(j = 0; j < size; j++) {
						SortData data = (SortData)list.get(j);
						if(data.key==null) break;
						if(keyComparator.compare(sd.key, data.key) * sortMask < 0) break;
					}
					if(j == size) list.add(sd);
					else list.add(j, sd);
				}
			} else 
				list.add(sd);
		}
		// sort SortData by index, execute following code only when sort flag is false
		if(!sort) Collections.sort(list, comparator);
		// finally, we set the fixed control's position
		//    first process controls which is adhere to other control
		int size = adhereToControls.size();
		for(int i = 0; i < size; i++) {
			SortData sd = (SortData)adhereToControls.get(i);
			int listSize = list.size();
			for(int j = 0; j < listSize; j++) {
				Control child = children[((SortData)list.get(j)).childIndex];
				if(child == sd.control) {
					// check adhere before or after
					if(sd.beforeAdheredControl) 
						list.add(j++, sd);
					else
						list.add(++j, sd);
				}
			}
		}
		// then process controls which is adhere to index
		//    we need sort it in order to insert smaller adhered index first
		Collections.sort(adhereToIndexs, comparator);
		size = adhereToIndexs.size();
		while(adhereToIndexs.size() > 0 && ((SortData)adhereToIndexs.get(0)).index <= list.size()) {
			for(int i = 0; i < size; i++) {
				SortData sd = (SortData)adhereToIndexs.get(i);
				if(sd.index == -1) 
					list.add(sd);
				else if(sd.index <= list.size())
					list.add(sd.index, sd);
				adhereToIndexs.remove(i--);				
				size--;
			}			
		}
		// add remaining SortData to the last
		size = adhereToIndexs.size();
		for(int i = 0; i < size; i++)
			list.add(adhereToIndexs.get(i));
	}
	
	// initialize and calculate some layout variables
	private void calculateListDimensions(Composite composite, boolean flushCache) {
		// grow columns and rows array
		int i;
		HashMap growCols = new HashMap();
		HashMap growRows = new HashMap();
		// initialize max height and max width array and set them zero
		Control[] children = composite.getChildren();
		colMaxWidth = new int[numColumns];
		for(i = 0; i < colMaxWidth.length; i++) colMaxWidth[i] = 0;
		rowMaxHeight = new int[(children.length - 1) / numColumns + 1];
		for(i = 0; i < rowMaxHeight.length; i++) rowMaxHeight[i] = 0;
		int row = 0, col = 0;
		// calculate max height of row and max width of col
		for(i = 0; i < list.size(); i++) {
			SortData sd = (SortData)list.get(i);
			Control child = children[sd.childIndex];
			Point childSize = child.computeSize(sd.widthHint, sd.heightHint, flushCache);
			colMaxWidth[col] = Math.max(colMaxWidth[col], childSize.x);
			rowMaxHeight[row] = Math.max(rowMaxHeight[row], childSize.y);
			// check whether it can expand
			if(sd.grabExcessHorizontalSpace) growCols.put(new Integer(col), null);
			if(sd.grabExcessVerticalSpace) growRows.put(new Integer(row), null);
			// modify row and col
			if(++col >= numColumns) {
				col = 0;
				row++;
			}
		}
		
		// initialize expandable rows and cols array
		expandableCols = new int[growCols.size()];
		expandableRows = new int[growRows.size()];
		Iterator iter = growCols.keySet().iterator();
		i = 0;
		while(iter.hasNext())
			expandableCols[i++] = ((Integer)iter.next()).intValue();
		iter = growRows.keySet().iterator();
		i = 0;
		while(iter.hasNext())
			expandableRows[i++] = ((Integer)iter.next()).intValue();
		
		// check whether columns have equal width, if so, reset max width
		if(makeColumnsEqualWidth) {
			int maxWidth = 0;
			for(i = 0; i < colMaxWidth.length; i++)
				maxWidth = Math.max(maxWidth, colMaxWidth[i]);
			for(i = 0; i < colMaxWidth.length; i++)
				colMaxWidth[i] = maxWidth;
		}
		
		// check whether rows have equal height, if so, reset max height
		if(makeRowsEqualHeight) {
			int maxHeight = 0;
			for(i = 0; i < rowMaxHeight.length; i++)
				maxHeight = Math.max(maxHeight, rowMaxHeight[i]);
			for(i = 0; i < rowMaxHeight.length; i++)
				rowMaxHeight[i] = maxHeight;
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.swt.widgets.Layout#computeSize(org.eclipse.swt.widgets.Composite, int, int, boolean)
	 */
	protected Point computeSize(Composite composite, int wHint, int hHint, boolean flushCache) {
		// if no children, return 0, 0
		Control children[] = composite.getChildren();
		if(children.length == 0)
			return new Point(0, 0);
		// compute client area size
		return computeLayoutSize(composite, wHint, hHint, flushCache);
	}

	/**
	 * compute client area size
	 * @param composite
	 * @param wHint
	 * @param hHint
	 * @param flushCache
	 * @return client area size
	 */
	private Point computeLayoutSize(Composite composite, int wHint, int hHint, boolean flushCache) {
		// if layout data array is empty, create it
		if(flushCache || colMaxWidth == null || rowMaxHeight == null) {
			list.removeAllElements();
			createList(composite);
			calculateListDimensions(composite, flushCache);
		}
		// compute the client area size, it's simple, what we need to do is
		//    adding all height and width plus margin and spacing
		int totalWidth = 0, totalHeight = 0;
		for(int i = 0; i < colMaxWidth.length; i++)
			totalWidth += colMaxWidth[i];
		totalWidth += (numColumns - 1) * horizontalSpacing + 2 * marginWidth;
		for(int i = 0; i < rowMaxHeight.length; i++)
			totalHeight += rowMaxHeight[i];
		totalHeight += (rowMaxHeight.length - 1) * verticalSpacing + 2 * marginHeight;
		
		// if wHint or hHint is not equal -1, use it
		if(wHint != -1)
			totalWidth = wHint;
		if(hHint != -1)
			totalHeight = hHint;
		return new Point(totalWidth, totalHeight);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.swt.widgets.Layout#layout(org.eclipse.swt.widgets.Composite, boolean)
	 */
	protected void layout(Composite composite, boolean flushCache) {
		int excess, horizontalExcess, horizontalRemainder, verticalExcess, verticalRemainder;
		// set sort mask
		sortMask = ascendent?1:-1;
		// if there is no children, do not layout
		Control[] children = composite.getChildren();
		if(children.length == 0)
			return;
		
		// compute client area size
		Point compositeSize = computeSize(composite, -1, -1, flushCache);
		
		// check whether columns have equal width, if so, reset max width
		//    we also need reset horizontal excess and remainder
		//    then we add excess to columns, the remainder is added to
		//    the last column
		if(makeColumnsEqualWidth) {
			excess = composite.getClientArea().width - compositeSize.x;
			horizontalExcess = excess / numColumns;
			horizontalRemainder = excess % numColumns;
			
			int maxWidth = 0;
			for(int i = 0; i < numColumns; i++)
				maxWidth = Math.max(maxWidth, colMaxWidth[i]);
			for(int i = 0; i < numColumns; i++)
				colMaxWidth[i] = maxWidth + horizontalExcess;
			colMaxWidth[numColumns - 1] += horizontalRemainder;			
		} else {
			if(expandableCols.length > 0) {
				excess = composite.getClientArea().width - compositeSize.x;
				horizontalExcess = excess / expandableCols.length;
				horizontalRemainder = excess % expandableCols.length;
				
				for(int i = 0; i < expandableCols.length; i++)
					colMaxWidth[expandableCols[i]] += horizontalExcess;
				colMaxWidth[colMaxWidth.length - 1] += horizontalRemainder;				
			}
		}
		
		// check whether rows have equal height, if so, reset max height
		//    we also need reset vertical excess and remainder
		//    then we add excess to rows, the remainder is added to the 
		//    last row
		if(makeRowsEqualHeight) {
			excess = composite.getClientArea().height - compositeSize.y;
			verticalExcess = excess / rowMaxHeight.length;
			verticalRemainder = excess % rowMaxHeight.length;
			
			int maxHeight = 0;
			for(int i = 0; i < rowMaxHeight.length; i++)
				maxHeight = Math.max(maxHeight, rowMaxHeight[i]);
			for(int i = 0; i < rowMaxHeight.length; i++)
				rowMaxHeight[i] = maxHeight + verticalExcess;
			rowMaxHeight[rowMaxHeight.length - 1] += verticalRemainder;
		} else {
			if(expandableRows.length > 0) {
				excess = composite.getClientArea().height - compositeSize.y;
				verticalExcess = excess / expandableRows.length;
				verticalRemainder = excess % expandableRows.length;
				
				for(int i = 0; i < expandableRows.length; i++)
					rowMaxHeight[expandableRows[i]] += verticalExcess;			
				rowMaxHeight[rowMaxHeight.length - 1] += verticalRemainder;				
			}
		}
		
		// begin set control's bound
		int size = list.size();
		int x = marginWidth, y = marginHeight, row = 0, col = 0;
		for(int i = 0; i < size; i++) {
			SortData sd = (SortData)list.get(i);
			Control child = children[sd.childIndex];
			Point childSize = child.computeSize(sd.widthHint, sd.heightHint, flushCache);
			Rectangle childBound = new Rectangle(0, 0, 0, 0);
			// set child x coordinate
			if(sd.horizontalAlignment==SortData.BEGINNING)
				childBound.x = x + sd.horizontalIndent;
			else if(sd.horizontalAlignment==SortData.CENTER)
				childBound.x = x + (colMaxWidth[col] - childSize.x) / 2 + sd.horizontalIndent;
			else if(sd.horizontalAlignment==SortData.END)
				childBound.x = x + colMaxWidth[col] - childSize.x + sd.horizontalIndent;
			else
				childBound.x = x;
			// set child y coordinate
			if(sd.verticalAlignment==SortData.BEGINNING)
				childBound.y = y + sd.verticalIndent;
			else if(sd.verticalAlignment==SortData.CENTER)
				childBound.y = y + (rowMaxHeight[row] - childSize.y) / 2 + sd.verticalIndent;
			else if(sd.verticalAlignment==SortData.END)
				childBound.y = y + rowMaxHeight[row] - childSize.y + sd.verticalIndent;
			else
				childBound.y = y;
			// set child width
			if(sd.horizontalAlignment==SortData.FILL)
				childBound.width = colMaxWidth[col];
			else
				childBound.width = childSize.x;
			// set child height
			if(sd.verticalAlignment==SortData.FILL)
				childBound.height = rowMaxHeight[row];
			else
				childBound.height = childSize.y;
			// modify col and row
			if(++col >= numColumns) {
				col = 0;
				row++;
				// modify x, y
				x = marginWidth;
				y += rowMaxHeight[row - 1] + verticalSpacing; 
			} else {
				// modify x, y
				x += colMaxWidth[col - 1] + horizontalSpacing;			
			}
			// set child bound
			child.setBounds(childBound);
		}
	}
}
