/*
* Shutter - Friend List Component
*
* Copyright (C) 2004 luma <stubma@163.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
package edu.tsinghua.swt.models;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import edu.tsinghua.swt.events.*;
import edu.tsinghua.swt.events.ModelListener;

/**
 * An array-based model implementation, intended to subclassed.
 * It use array to store node, so the access is sequential
 * @author luma@smth
 */
public abstract class AbstractArrayModel implements IModel, PropertyListener {
	// model listeners
	protected ArrayList modelListeners;
	// item property listeners
	protected ArrayList propListeners;
	// items
	protected ArrayList nodeArray;

	/**
	 * constructor
	 */
	public AbstractArrayModel() {
		modelListeners = new ArrayList();
		propListeners = new ArrayList();
		nodeArray = new ArrayList();
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.IModel#find(edu.tsinghua.swt.models.IMatcher)
	 */
	public List find(IMatcher matcher) {
	    if(matcher == null) return null;
	    List ret = new ArrayList();
	    Iterator iter = nodeArray.iterator();
	    while(iter.hasNext()) {
	    	INode node = (INode)iter.next();
	    	if(matcher.match(node))
	    		ret.add(node);
	    }
	    return ret;
	}
	
	/**
	 * зNodeǵindex
	 * @param matcher IMatcher
	 * @return
	 */
	protected int[] findIndex(IMatcher matcher) {
		if(matcher == null) return null;
		int[] indices = new int[nodeArray.size()];
		int size = indices.length;
		int cur = 0;
		for(int i = 0; i < size; i++) {
			INode node = (INode)nodeArray.get(i);
			if(matcher.match(node))
				indices[cur++] = i;
		}
		int[] ret = new int[cur];
		System.arraycopy(indices, 0, ret, 0, cur);
		return ret;
	}
    
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.IModel#remove(java.lang.String, java.lang.Object)
	 */
	public boolean remove(INode node) {
		return nodeArray.remove(node);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.shutter.IModel#addItem(edu.tsinghua.shutter.GeneralItem)
	 */
	public void add(INode node) {
		nodeArray.add(node);
		node.addPropertyListener(this);
	}

	/**
	 * remove a node according its index
	 * @param index node index
	 * @return if remove success, return INode removed; otherwise return null
	 */
	public INode remove(int index) {
		if(index < 0 || index >= nodeArray.size()) return null;
		INode ret = (INode)nodeArray.remove(index);
		if(ret != null)
			ret.removePropertyListener(this);
		return ret;
	}
	
	/**
	 * insert a node with specified index
	 * @param index insert index
	 * @param node inserted index
	 * @return true if insert is successful, false if the index is out of bound
	 */
	public boolean insert(int index, INode node) {
		if(index < 0 || index > nodeArray.size()) return false;
		nodeArray.add(index, node);
		node.addPropertyListener(this);
		return true;
	}
	
	/**
	 * get a node according its index
	 * @param index node index
	 * @return node object, null if index is outbound
	 */
	public INode get(int index) {
		if(index < 0 || index >= nodeArray.size()) return null;
		return (INode)nodeArray.get(index);
	}
	
	/**
	 * return a node's index
	 * @param node node object
	 * @return node index, -1 if node isn't exist
	 */
	public int indexOf(Object node) {
		return nodeArray.indexOf(node);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.shutter.IModel#addGeneralModelListener(edu.tsinghua.shutter.ModelListener)
	 */
	public void addModelListener(ModelListener listener) {
		modelListeners.add(listener);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.shutter.IModel#removeGeneralModelListener(edu.tsinghua.shutter.ModelListener)
	 */
	public void removeModelListener(ModelListener listener) {
		modelListeners.remove(listener);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.IModel#addPropertyListener(edu.tsinghua.swt.models.PropertyListener)
	 */
	public void addPropertyListener(PropertyListener listener) {
		propListeners.add(listener);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.IModel#removePropertyListener(edu.tsinghua.swt.models.PropertyListener)
	 */
	public void removePropertyListener(PropertyListener listener) {
		propListeners.remove(listener);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.events.PropertyListener#propertyAdded(edu.tsinghua.swt.events.PropertyEvent)
	 */
	public void propertyAdded(PropertyEvent e) {
		fireItemPropertyAddedEvent(e);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.events.PropertyListener#propertyModified(edu.tsinghua.swt.events.PropertyEvent)
	 */
	public void propertyModified(PropertyEvent e) {
		fireItemPropertyModifiedEvent(e);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.events.PropertyListener#propertyRemoved(edu.tsinghua.swt.events.PropertyEvent)
	 */
	public void propertyRemoved(PropertyEvent e) {
		fireItemPropertyRemovedEvent(e);
	}	
	
	/**
	 * trigger item property add event
	 * @param e
	 */
	protected void fireItemPropertyAddedEvent(PropertyEvent e) {
		int count = propListeners.size();
		for(int i=0;i<count;i++)
			((PropertyListener)propListeners.get(i)).propertyAdded(e);
	}
	
	/**
	 * trigger item property remove event
	 * @param e
	 */
	protected void fireItemPropertyRemovedEvent(PropertyEvent e) {
		int count = propListeners.size();
		for(int i=0;i<count;i++)
			((PropertyListener)propListeners.get(i)).propertyRemoved(e);
	}
	
	/**
	 * trigger item property add event
	 * @param e
	 */
	protected void fireItemPropertyModifiedEvent(PropertyEvent e) {
		int count = propListeners.size();
		for(int i=0;i<count;i++)
			((PropertyListener)propListeners.get(i)).propertyModified(e);
	}		
}
