/**
* Shutter - Friend List Component
*
* Copyright (C) 2004 luma <stubma@163.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
package edu.tsinghua.swt.models;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import edu.tsinghua.swt.events.ShutterModelEvent;
import edu.tsinghua.swt.events.ShutterModelListener;

/**
 * Abstract shutter model, intended to subclass by your implementation
 * @author luma@smth
 */
public abstract class AbstractShutterModel extends AbstractArrayModel implements ShutterModel {
	// model in each tab
	protected ArrayList tabModels;
	
	/**
	 * default constructor
	 */
	public AbstractShutterModel() {
		super();
		tabModels = new ArrayList();
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.shutter.ShutterModel#getItemCount(int)
	 */
	public int getItemCount(int tabIndex) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size()) return 0;
		return ((ScrollListModel)tabModels.get(tabIndex)).getItemCount();
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.shutter.ShutterModel#getTabCount()
	 */
	public int getTabCount() {
		return nodeArray.size();
	}

	/**
	 * fire a tab added event
	 * @param tabIndex index of tab which is added
	 */
	protected void fireTabAddedEvent(int tabIndex) {
		Iterator iter = modelListeners.iterator();
		while(iter.hasNext())
			((ShutterModelListener)iter.next()).modelAdded(new ShutterModelEvent(this, tabIndex, -1));
	}
	
	/**
	 * fire a tab removed event
	 * @param tabIndex index of tab which is removed
	 */
	protected void fireTabRemovedEvent(int tabIndex) {
		Iterator iter = modelListeners.iterator();
		while(iter.hasNext())
			((ShutterModelListener)iter.next()).modelRemoved(new ShutterModelEvent(this, tabIndex, -1));
	}
	
	/**
	 * fire an item added event
	 * @param tabIndex index of tab in which item is added
	 * @param itemIndex index of item which is added
	 */
	protected void fireItemAddedEvent(int tabIndex, int itemIndex) {
		Iterator iter = modelListeners.iterator();
		while(iter.hasNext())
			((ShutterModelListener)iter.next()).itemAdded(new ShutterModelEvent(this, tabIndex, itemIndex));
	}
	
	/**
	 * fire an item removed event
	 * @param tabIndex index of tab in which item is removed
	 * @param itemIndex index of item which is removed
	 */
	protected void fireItemRemovedEvent(int tabIndex, int itemIndex) {
		Iterator iter = modelListeners.iterator();
		while(iter.hasNext())
			((ShutterModelListener)iter.next()).itemRemoved(new ShutterModelEvent(this, tabIndex, itemIndex));
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#addTab(edu.tsinghua.swt.models.INode)
	 */
	public void addTab(INode tab) {
		super.add(tab);
		ScrollListModel tabModel = new DefaultScrollListModel();
		tabModels.add(tabModel);
		fireTabAddedEvent(nodeArray.size() - 1);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#insertTab(int, edu.tsinghua.swt.models.INode)
	 */
	public boolean insertTab(int index, INode tab) {
		boolean succeed = super.insert(index, tab);
		if(succeed) {
			ScrollListModel tabModel = new DefaultScrollListModel();
			tabModels.add(index, tabModel);
			fireTabAddedEvent(index);
		}
		return succeed;
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#removeTab(int)
	 */
	public INode removeTab(int index) {
		// if remove succeed, trigger event
		INode ret = (INode)super.remove(index);
		if(ret != null) {
			tabModels.remove(index);
			fireTabRemovedEvent(index);
		}
		return ret;
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#getTab(int)
	 */
	public INode getTab(int tabIndex) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size()) return null;
		return (INode)nodeArray.get(tabIndex);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#addItem(int, edu.tsinghua.swt.models.INode)
	 */
	public void addItem(int tabIndex, INode item) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size()) return;
		((ScrollListModel)tabModels.get(tabIndex)).addItem(item);
		fireItemAddedEvent(tabIndex, getItemCount(tabIndex) - 1);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#removeItem(int, int)
	 */
	public INode removeItem(int tabIndex, int itemIndex) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size() || itemIndex < 0 || itemIndex >= getItemCount(tabIndex)) return null;
		INode ret = ((ScrollListModel)tabModels.get(tabIndex)).removeItem(itemIndex);
		if(ret != null)
			fireItemRemovedEvent(tabIndex, itemIndex);
		return ret;
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#getItem(int, int)
	 */
	public INode getItem(int tabIndex, int itemIndex) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size() || itemIndex < 0 || itemIndex >= getItemCount(tabIndex)) return null;
		return (INode)((ScrollListModel)tabModels.get(tabIndex)).getItem(itemIndex);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#insertItem(int, int, edu.tsinghua.swt.models.INode)
	 */
	public boolean insertItem(int tabIndex, int itemIndex, INode item) {
		if(tabIndex < 0 || tabIndex >= nodeArray.size()) return false;
		boolean succeed = ((ScrollListModel)tabModels.get(tabIndex)).insertItem(itemIndex, item);
		if(succeed)
			fireItemAddedEvent(tabIndex, itemIndex);
		return succeed;
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#getTabModel(int)
	 */
	public ScrollListModel getTabModel(int tabIndex) {
		if(tabIndex < 0 || tabIndex >= tabModels.size()) return null;
		return (ScrollListModel)tabModels.get(tabIndex);
	}

	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#setTabModel(int, edu.tsinghua.swt.models.ScrollListModel)
	 */
	public void setTabModel(int tabIndex, ScrollListModel model) {
		if(tabIndex < 0 || tabIndex >= tabModels.size()) return;
		tabModels.add(tabIndex, model);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#findItem(edu.tsinghua.swt.models.IMatcher)
	 */
	public List findItem(IMatcher matcher) {
		int size = tabModels.size();
		List ret = new ArrayList();
		for(int i = 0; i < size; i++) {
			List l = ((ScrollListModel)tabModels.get(i)).find(matcher);
			if(l != null)
				ret.addAll(l);
		}
		return ret;
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#getAllItemCount()
	 */
	public int getAllItemCount() {
		int total = 0;
		int size = tabModels.size();
		for(int i = 0; i < size; i++)
			total += ((ScrollListModel)tabModels.get(i)).getItemCount();
		return total;
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#findItemIndex(int, edu.tsinghua.swt.models.IMatcher)
	 */
	public int[] findItemIndex(int tabIndex, IMatcher matcher) {
		if(tabIndex < 0 || tabIndex > nodeArray.size()) return null;
		return ((ScrollListModel)tabModels.get(tabIndex)).findItemIndex(matcher);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#findTabIndex(edu.tsinghua.swt.models.IMatcher)
	 */
	public int[] findTabIndex(IMatcher matcher) {
		return super.findIndex(matcher);
	}
	
	/* (non-Javadoc)
	 * @see edu.tsinghua.swt.models.ShutterModel#indexOfItem(int, java.lang.Object)
	 */
	public int indexOfItem(int tabIndex, Object item) {
		if(tabIndex < 0 || tabIndex > nodeArray.size()) return -1;
		return ((ScrollListModel)tabModels.get(tabIndex)).indexOf(item);
	}
}
